from __future__ import print_function
import os
import pickle
import time
import re

import matplotlib.pyplot as plt
import numpy as np
from music21 import *

import DT
import chord_distributions as cd

"""
This is one of the main routines for reading raw progressions in various corpora -- the almost-complete works of Ockeghem, Josquin ("Josquincomplete"), Palestrina ("Palraw"), and Victoria ("VictoriaRaw").

The code is a little messy because it is redundant; the earlier version gathered progressions directly; later, I created a dict called fullProgs that allows for more flexibility; it gets converted into a series 
of chord progressions of arbitrary length using chord_progressions; these can be easily manipulated.  The advantage of this method is that it allows you to control various aspects of the conversion without
reanalyzing the data.


"""

_THISFILE = 'PROGRESSIONS'

pieceCounts = {}
pieceList = []
massFiles = {}
finalChords = {}
keySigs = {}

def gather_stats(source = 'BACH', fileName = '', printOut = False):
	global progs, piece, globalSource, locNames, globalPrint, bassProgs, accidentalHisto, fullProgs, durationHisto
	startT = time.time()
	globalPrint = printOut
	globalSource = source.upper()
	progs = {}
	locNames = {}
	bassProgs = {}
	fullProgs = {}
	accidentalHisto = {}
	durationHisto = {}
	if globalSource not in DT._PATHS:
		print("Source not found!")
		return
	scorePath = DT._PATHS[globalSource][1]
	files = sorted(os.listdir(scorePath))
	counter = 1
	for myFile in files:
		if myFile[-3:] in ['xml', 'krn', 'mxl']:
			if fileName and myFile != fileName: continue
			print('Analyzing', myFile)
			piece = converter.parse(scorePath + myFile)
			if 'Opus' in piece.classes:
				bigOpus = piece
				for opusSection, piece in enumerate(bigOpus):
					if 'Score' not in piece.classes: continue
					print("  multi-score file, section:", opusSection + 1, end=' ')
					try:
						dochorale(counter)
						locNames[counter] = str(opusSection) + '***}' + myFile
						counter += 1
					except:
						print("ERROR: ", tempName, sys.exc_info()[0])
			else:
				dopiece(counter)
				locNames[counter] = myFile
				counter += 1
			if counter % 50 == 0:
				save()
	if not fileName:
		save()
	minutes = (time.time() - startT)/60.
	print('Total analysis took', int(minutes), 'minutes and %.0f seconds' % ((minutes - int(minutes)) * 60))

def save():
	with open(DT._DATAPATH + _THISFILE + globalSource +'.pkl', 'wb') as outputFile:
		pickle.dump(progs, outputFile)
		pickle.dump(locNames, outputFile)
		pickle.dump(bassProgs, outputFile)
		pickle.dump(accidentalHisto, outputFile)
		pickle.dump(fullProgs, outputFile)
		pickle.dump(durationHisto, outputFile)

def load(source = 'BACH', loadCD = True):
	global masterProgs, globalSource, locNames, bassProgs, accidentalHisto, fullProgs, pieceList, durationHisto, locNumbers
	try:
		globalSource = source.upper()
		with open(DT._DATAPATH + _THISFILE + globalSource +'.pkl', 'rb') as inputFile:
			masterProgs = pickle.load(inputFile)
			locNames = pickle.load(inputFile)
			pieceList = []
			bassProgs = pickle.load(inputFile)
			accidentalHisto = pickle.load(inputFile)
			fullProgs = pickle.load(inputFile)
			durationHisto = pickle.load(inputFile)
		get_data()
		locNumbers = {v:k for k, v in locNames.items()}
		if loadCD:
			cd.load(globalSource)
		return True
	except:
		return False

def get_data(pieceList = [], skipMax = 2):
	global progs, masterProgs, globalSource, locNames, rootProgs, progs2, rootProgs2
	if not pieceList:
		if globalSource not in pieceCounts:
			pieceList = masterProgs.keys()
		else:
			pieceList = pieceCounts[globalSource]
	progs = {}
	for i in pieceList:
		for j in masterProgs[i]:
			if j <= skipMax:
				combine_dicts(progs, masterProgs[i][j])
	rootProgs = {}
	rootProgs2 = {}
	progs2 = {}
	for k in progs:
		newKey = k[:4] + str(root_motion(k))
		rootProgs[newKey] = progs[k]
		rootProgs2[newKey[:2]+newKey[4:]] = rootProgs2.setdefault(newKey[:2]+newKey[4:], 0) + progs[k]
		progs2[k[:2]+k[4:]] = progs2.setdefault(k[:2]+k[4:], 0) + progs[k]
		
def remove_incomplete(skipMax = 2):		# add weights
	global progs, masterProgs, globalSource, locNames, rootProgs, progs2, rootProgs2, bassProgs, pieceList
	if not pieceList:
		pieceList = fullProgs.keys()
	progs = {}
	progs2 = {}
	tempBassProgs = {}
	for i in pieceList:
		lastChord = False
		skippedChords = 0
		tempBassProgs[i] = {}
		for o, sym in fullProgs[i]:
			if good_chord(sym):
				if lastChord and skippedChords <= skipMax and lastChord[:2] != sym[:2]:
					key = combine_symbols(lastChord, sym) + str(get_interval_from_letters(lastChord, sym))
					if skippedChords not in tempBassProgs[i]:
						tempBassProgs[i][skippedChords] = {}
					tempBassProgs[i][skippedChords][key] = tempBassProgs[i][skippedChords].get(key, 0) + 1
					key = key[1:]
					progs[key] = progs.get(key, 0) + 1
					if key[-2] == '-':
						key = key[:2] + key[-2:]
					else:
						key = key[:2] + key[-1]
					progs2[key] = progs2.get(key, 0) + 1
				lastChord = sym
				skippedChords = 0
			else:
				skippedChords += 1
	bassProgs = tempBassProgs


def chord_progressions(length = 4, skipMax = 2, includeIncomplete = False, skipIncomplete = False, dontCountSubsets = True):		# add weights
	global longProgs, pieceList
	if not pieceList:
		pieceList = fullProgs.keys()
	longProgs = {}
	for i in pieceList:
		lastChord = False
		skippedChords = 0
		goodChords = []
		for j, l in enumerate(fullProgs[i]):
			o, sym = l
			if good_chord(sym, includeIncomplete = includeIncomplete):
				if not goodChords or skippedChords > skipMax:
					goodChords = [sym]
					skippedChords = 0
				else:
					if goodChords[-1] != sym:
						goodChords.append(sym)
						skippedChords = 0
				if len(goodChords) == length:
					key = ' '.join(goodChords)
					longProgs.setdefault(key, []).append([i, o])
					goodChords = goodChords[1:]
			else:
				if skipIncomplete or sym.count('?') > 0 or (dontCountSubsets and len(goodChords) > 0 and subset(sym, goodChords[-1])):
					skippedChords += 1
				else:
					skippedChords += 100000

lastChoraleNo = False
globalShowIndex = None
lastShowChord = False

def show(rnProg, index = None, start = 1, end = 1):
	global output, locNames, globalSource, longProgs, lastChoraleNo, chorale, globalShowIndex, lastShowChord
	if rnProg != lastShowChord:
		lastShowChord = rnProg
		if index == None:
			globalShowIndex = None
	output = stream.Score()
	tempOutput = stream.Score()
	locList = longProgs.get(rnProg, [])
	if index == None:
		if globalShowIndex == None:
			index = 0
			globalShowIndex = 0
		else:
			index = globalShowIndex + 1
			globalShowIndex += 1
	else:
		globalShowIndex = index
	if not locList:
		print('Not found')
		globalShowIndex = None
		return
	if len(locList) <= globalShowIndex:
		print("no more to get")
		globalShowIndex = None
		return
	pieceNumber, myOffset = locList[index]
	if globalSource == 'PALRAW':
		s = cd.palestrinaData.get(locNames[pieceNumber], '')
		s = ' '.join([str(x) for x in s])
	else:
		s = ''
	print('getting', locNames[pieceNumber], "index", index, "of", len(locList))
	if pieceNumber != lastChoraleNo:
		choraleNo = pieceNumber
		lastChoraleNo = pieceNumber
		chorale = converter.parse(DT._PATHS[globalSource][1] + locNames[pieceNumber])
		if 'Opus' in chorale.classes:
			print('MULTIMOVEMENT WORK!')
			x = multimovement_work(pieceNumber)
			if not x:
				print("  can't get piece number")
				return
			subpiece = x.index(pieceNumber)
			chorale = chorale.scores[subpiece]
			s += 'subpiece ' + str(subpiece)
	mNum = DT.get_measure_number_and_beat_from_offset(chorale, myOffset)[0]
	measureList = [max([mNum - start, 0]), mNum + end]
	header = str(locNames[pieceNumber])+ ' ' + s + ' ' + 'm'+str(measureList[0])
	output = DT.get_measures(chorale, int(measureList[0]), int(measureList[1]))
	if output:
		try:
			DT.insert_text(output.parts[0][0], header, 55, 0)
		except:
			DT.insert_text(output.parts[1][0], header, 55, 0)
		output.show()
		return output

def multimovement_work(i):
	mList = [x for x in locNames if locNames[x] == locNames[i]]
	if len(mList) == 1:
		return False
	else:
		return sorted(mList)

def good_chord(chord, includeIncomplete = False):
	if includeIncomplete:
		if chord.count("?") == 0:
			return True
		return False
	if len(chord) < 2 or chord[1] not in 'RTF':
		if len(chord) > 2 and chord[1] in ['-', '#'] and chord[2] in 'RTF':
			return True
		return False
	return True

def combine_symbols(s1, s2):
	return s1[:2] + s2[1] + s1[2] + s2[2]

globalLetterNames = 'ABCDEFG'

def get_interval_from_letters(l1, l2): 
	if l1[0] not in globalLetterNames or l2[0] not in globalLetterNames:
		return None
	theInt = (globalLetterNames.index(l2[0]) - globalLetterNames.index(l1[0])) % 7
	if theInt > 3: theInt -= 7
	return theInt

def get_piece(locNumber, show = True):
	global massFiles
	if type(locNumber) is str:
		locNumber = locNumbers[locNumber]	# input a piece name
	if globalSource == 'PALRAW' and not massFiles:
		get_piece_files(printOut = False)
		last_chord_by_piece(printOut = False)
	scorePath = DT._PATHS[globalSource][1]
	if type(locNumber) is str:
		piece = converter.parse(scorePath + locNumber)
		piece.show()
		return
	locName = locNames[locNumber]
	files = os.listdir(scorePath)
	targetFile = locName
	if targetFile in files:
		if globalSource == 'PALRAW':
			data = cd.palestrinaData.get(targetFile, '')
			print(targetFile, data)
			if data:
				print(sorted(massFiles[data[0]], key = sort_mass_movements))
				print([x[1] for x in sorted(finalChords[data[0]], key = lambda x: sort_mass_movements(x[0]))])
				print('key signature', keySigs[data[0]][0])
		else:
			print(targetFile)
		piece = converter.parse(scorePath + targetFile)
		if show:
			piece.show()
			return
		else: 
			return piece

def combine_dicts(outputDict, *otherDicts):
	for newDict in otherDicts:
		for k, v in newDict.items():
			outputDict[k] = outputDict.setdefault(k, 0) + v

def root_motion(s):
	tempDict = {'R': 0, 'T': 2, 'F': 4, '6': 2, '3': 0, '5': 0, '1': 0}
	bassMotion = int(s[4:])
	i = (bassMotion + tempDict[s[0]] - tempDict[s[1]]) % 7
	if i > 3: i = i - 7
	return i		

fbDict = {(0, 2, 4): 'R', (0, 2, 5): 'T', (0, 3, 5): 'F', (0, 2): '3', (0, 4): '5', (0, 5): '6', (0,): '1'}
nfDict = {(0, 4, 7): 'M', (0, 3, 7): 'm', (0, 4, 8): 'A', (0, 3, 6): 'd', (0, 5): 'P', (0, 4): 'M', (0, 3): 'm', (0, 8): 'm', (0, 9): 'M', (0,): 'P'}

def dopiece(pieceNumber):	
	global progs, piece, consonantO, globalPrint, accidentalHisto, fullProgs
	score = [piece.parts[i].flat.getElementsByClass(['Note', 'Chord']).stream() for i in range(len(piece.parts))]
	newO = sorted(list(set([n.offset for p in score for n in p])))
	consonantO = {}
	progs[pieceNumber] = {}
	bassProgs[pieceNumber] = {}
	accidentalHisto[pieceNumber] = {}
	durationHisto[pieceNumber] = {}
	fullProgs[pieceNumber] = []
	for o in newO:
		myMidi = []
		myNotes = []
		for p in score:
			tempNotes = p.getElementsByOffset(o + .01, o + .02, mustBeginInSpan = False)
			if tempNotes:
				for n in tempNotes:
					if 'Note' in n.classes:
						myNotes.append(n)
					elif 'Chord' in n.classes:
						for n1 in n.pitches:
							myNotes.append(note.Note(n1))
		diatonicNotes = sorted([DT.note_to_scalarMIDIdegree(x) for x in myNotes])
		fbSetClass = tuple(sorted(list(set([(x - diatonicNotes[0]) % 7 for x in diatonicNotes]))))
		nForm = tuple(DT.normal_form([x.pitch.midi for x in myNotes], removeDuplicates = True))
		for n in myNotes:				# keep track of accidentals, for mode identification
			if n.offset == o:
				accidentalHisto[pieceNumber][n.name] = accidentalHisto[pieceNumber].get(n.name, 0) + 1
				durationHisto[pieceNumber][n.name] = durationHisto[pieceNumber].get(n.name, 0) + n.quarterLength
		if not diatonicNotes or len(diatonicNotes) == 1: continue
		if fbSetClass in fbDict:
			consonantO[o] = [min(myNotes), fbDict[fbSetClass], nfDict.get(nForm, '?')]
			if globalPrint: print(DT.get_measure_number_and_beat_from_offset(piece, o), consonantO[o])
	oList = sorted(consonantO.keys())
	
	for o in newO:
		if o in consonantO:
			theSymbol = consonantO[o][0].name + consonantO[o][1] + consonantO[o][2]
		else:
			theSymbol = '?'	
		fullProgs[pieceNumber].append([o, theSymbol])
		
	for i in range(len(oList) - 1):
		o = oList[i]
		nextO = oList[i+1]
		oIndex = newO.index(o)
		skippedChords = newO.index(nextO) - oIndex - 1
		if  skippedChords > 2: continue
		myInt = interval.notesToInterval(consonantO[o][0], consonantO[nextO][0])
		myInt = fix_int(myInt.generic.value) % 7
		if myInt > 3: myInt = myInt - 7
		theSymbol = consonantO[o][1] + consonantO[nextO][1] + consonantO[o][2] + consonantO[nextO][2] + str(myInt)
		#theSymbol = 'RTF'[consonantO[o][1]] + 'RTF'[consonantO[nextO][1]] + consonantO[o][2] + consonantO[nextO][2] + str(myInt)
		if theSymbol[0] == theSymbol[1] and theSymbol[2] == theSymbol[3] and myInt == 0:
			pass
		else:
			if skippedChords not in progs[pieceNumber]:
				progs[pieceNumber][skippedChords] = {}
				bassProgs[pieceNumber][skippedChords] = {}
			progs[pieceNumber][skippedChords][theSymbol] = progs[pieceNumber][skippedChords].setdefault(theSymbol, 0) + 1
			theSymbol = consonantO[o][0].name + theSymbol
			bassProgs[pieceNumber][skippedChords][theSymbol] = bassProgs[pieceNumber][skippedChords].setdefault(theSymbol, 0) + 1
		
def fix_int(i):
	if i > 0: return i -1
	return i + 1

def figured_bass_midi(myNotes):
	sortedNotes = sorted([x.pitch.midi for x in myNotes])
	tempList = sorted(list(set([(x - sortedNotes[0]) % 12 for x in sortedNotes])))
	return tempList
	#return [(x + sortedNotes[0]) % 12 for x in tempList]

allProgs = {}

def regex(patternStrings, useLongProgs = False, qualities = False, roots = False, printIt = True, pct = True, filterRare = 2, useSum = False, removeIncomplete = False):
	"""
	
	call chord_progressions() first to determine length of progression
	then call regex ... CRM means "bass is C, chord is in root position, triad quality is Major.  
		example: progressions.regex(".* .F. ATd GRM$")
	can call progressions.show("CTm GFM ATd GRM")
	
	"""
	global results, progs, rootProgs, progs2, rootProgs2, allProgs, bigSum
	if removeIncomplete:
		remove_incomplete()
	source = [[progs2, rootProgs2][int(roots)], [progs, rootProgs][int(roots)]][int(qualities)]
	if useLongProgs:
		source = longProgs
	results = {}
	theSums = {}
	if type(patternStrings) is str:
		patternStrings = [patternStrings] 
	if useLongProgs:
		bigSum = sum([len(x) for x in source])
	else:
		bigSum = sum(source.values())
	for prog in source:
		for patternString in patternStrings:
			if patternString not in results:
				results[patternString] = []
			if re.match(patternString, prog):
				if useLongProgs:
					results[patternString].append([prog, len(source[prog])])
				else:
					results[patternString].append([prog, source[prog]])
	for patternString in patternStrings:
		tempList = sorted(results[patternString], key=lambda x: -x[1])
		mySum = sum([x[1] for x in tempList])
		print(patternString, mySum, "%.1f of all progressions" % (100. * mySum/bigSum))
		if useSum:
			tempList = [[patternString, mySum * mySum/bigSum]]
		for pair in tempList: 
			if pct:
				p = (100. * pair[1]/mySum)
				#print pair[0], allProgs.setdefault(pair[0], [])
				allProgs[pair[0]] = allProgs.setdefault(pair[0], [])
				allProgs[pair[0]].append(p)
			if printIt:
				if not filterRare or p > filterRare:
					if pct:
						print(pair[0], int(round(p, 0)))
					else:
						print(pair[0], pair[1])

def doall(*args, **kwargs):
	global allProgs
	allProgs = {}
	for c in ['OCKEGHEM', 'JOSQUINCOMPLETE', 'PALRAW']:
		print('Composer ', c)
		load(c)
		regex(*args, **kwargs)
	print() 
	theKeys = [x for x in allProgs.keys() if len(allProgs[x]) > 2]
	k = sorted(theKeys, key = lambda x: -abs(allProgs[x][0] - allProgs[x][-1]))
	anomalies = []
	for theKey in k:
		print(theKey, ' '.join(["%.0f" % x for x in allProgs[theKey]]))
		data = allProgs[theKey]
		if (data[1] > data[0] and data[1] > data[2]) or (data[1] < data[0] and data[1] < data[2]):
			anomalies.append(theKey)
	print()
	print()
	print()
	print("ANOMALIES")
	for theKey in anomalies:
		print(theKey, ' '.join(["%.0f" % x for x in allProgs[theKey]]))
		

def root_data(targetNote = 'C', skipLength = 2, cutoff = 10, printOut = True):
	global output, pieceList
	output = {}
	for i in bassProgs:
		if pieceList and i not in pieceList: continue
		for j in bassProgs[i]:
			if j > skipLength: continue
			for s, count in bassProgs[i][j].items():
				ind = 1
				if s[1] in ['-', '#']:
					ind += 1
				theNote = s[:ind]
				theSym = s[ind:]
				theSym = theSym[:2] + theSym[4:]
				if theNote == targetNote:
					output[theSym] = output.get(theSym, 0) + count
	if printOut:
		DT.print_dict(output, pct = True, cutoff = cutoff)

testNotes = ['C', 'G', 'D', 'A', 'E', 'F', 'B-', 'B']

def popular_prog(tonicList = [0, 7, 2, 9, 4], targetSig = 0, top = 5, qualities = False):
	
	if type(targetSig) is int:
		targetSig = tuple([targetSig])
	for t in tonicList:
		totalSum = 0
		output = {}
		limit_pieces(tuple([t]), targetSig)
		for i in bassProgs:
			if pieceList and i not in pieceList: continue
			for j in range(2):
				for s, count in bassProgs[i][j].items():
					if not qualities:
						s = s.replace('M', '').replace('m', '')
					output[s] = output.get(s, 0) + count
		print(tuple([t]), targetSig)
		theKeys = sorted(output.keys(), key = lambda x: -output[x])
		normalize_dict(output)
		for i in range(top):
			print("  ", theKeys[i], output[theKeys[i]])	


def graph(targetProg = 'RR3', tonicList = [0, 7, 2, 9, 4], targetSig = 0, normalize = True, individualNormalize = False):
	
	if type(targetSig) is int:
		targetSig = tuple([targetSig])
	for t in tonicList:
		
		print(tuple([t]), targetSig)
		totalSum = 0
		
		limit_pieces(tuple([t]), targetSig)
		values = []
	
		for s in testNotes:
			root_data(s, printOut = False)
			if individualNormalize:
				normalize_dict(output)
			totalSum += sum(output.values())
				
			values.append(output.get(targetProg, 0))
		
		if normalize and not individualNormalize:
			values = [x*100./totalSum for x in values]
		plt.plot(values, label = t)
	
	plt.xlabel('Starting Bass Note')
	plt.ylabel('Percentage of all progressions starting with that bass that move by ' + targetProg)
	ticks = range(len(testNotes))
	plt.xticks(ticks, testNotes)
	myLeg = plt.legend(tuple([str(x) for x in tonicList]), loc=0)
	for i in range(len(myLeg.legendHandles)):
		myLeg.legendHandles[i].set_linewidth(2.0)
	plt.show()

#def distribution_graph(tonicList = [0, 7], targetSig = 0, bestChord = [(0, 4, 7), (7, 11, 2)])

def distribution_graph(tonicList = [2, 9], targetSig = 0, bestChord = [(2, 5, 9), (9, 0, 4)], equivalenceClasses = range(12), transposeToC = True):
	global myDict
	
	theRoots = [0, 2, 4, 5, 6, 7, 9, 10, 11] #[0, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11]
	noteNameDict = {0: 'C', 1: 'C#', 2: 'D', 4: 'E', 5: 'F', 7: 'G', 8: 'G#', 9: 'A', 10:'Bb', 11:'B', 6: 'F#'}
	
	if type(targetSig) is int:
		targetSig = tuple([targetSig])
		
	for tonicNumber, t in enumerate(tonicList):
		
		print(tuple([t]), targetSig)
		
		cd.limit_pieces(tuple([t]), targetSig, bestChord = bestChord[tonicNumber], transposeToC = transposeToC)
		
		myDict = {x[0]:x[1] for x in cd.average(equivalenceClasses = equivalenceClasses, subsets = False, transposeToC = transposeToC)}
		
		for firstInversion in [False, True]:
			
			values = []
	
			for s in theRoots:
				"""if transposeToC:
					s = (s + (targetSig*-7)) % 12"""
				if firstInversion:
					addresses = [tuple([(s - 8) % 12, (s-5) % 12, s % 12]), tuple([(s - 9) % 12, (s-5) % 12, s % 12]), tuple([(s - 9) % 12, (s-6) % 12, s % 12])]
				else:
					addresses = [tuple([s, (s+3) % 12, (s+7) % 12])]
				values.append(sum([myDict.get(x, 0) for x in addresses]))
			
			if not firstInversion:
				print("tonic", t, '\t'.join([str(round(x, 1)) for x in values]))
			plt.plot(values, label = str(t) + ['', '6'][int(firstInversion)])
	
	plt.xlabel('Root Note')
	plt.ylabel('Percentage of all sonorities with that root')
	ticks = range(len(theRoots))
	plt.xticks(ticks, [noteNameDict[x] for x in theRoots])
	legendTuple = []
	for x in tonicList:
		s = noteNameDict[x]
		legendTuple.append(s + ' center (root pos.)')
		legendTuple.append(s + ' center (first inv.)')
	myLeg = plt.legend(tuple(legendTuple), loc=0)
	for i in range(len(myLeg.legendHandles)):
		myLeg.legendHandles[i].set_linewidth(2.0)
	plt.show()

def tonal_type_graph(tonalTypes = [], popularChord = True, separateHighPart = True, transposeToC = True, equivalenceClasses = range(12)):
	global myDict
	
	theRoots = [0, 2, 4, 5, 6, 7, 9, 10, 11] #[0, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11]
	noteNameDict = {0: 'C', 1: 'C#', 2: 'D', 4: 'E', 5: 'F', 7: 'G', 8: 'G#', 9: 'A', 10:'Bb', 11:'B', 6: 'F#'}
	
	cd.tonal_types(popularChord = popularChord, separateHighPart = separateHighPart, printOut = False)
	
	for t in tonalTypes:
		
		cd.pieceNumbers = cd.tonalTypes[t]
		
		myDict = {x[0]:x[1] for x in cd.average(equivalenceClasses = equivalenceClasses, subsets = False, transposeToC = transposeToC)}
		
		for firstInversion in [False, True]:
			
			values = []
	
			for s in theRoots:
				"""if transposeToC:
					s = (s + (targetSig*-7)) % 12"""
				if firstInversion:
					addresses = [tuple([(s - 8) % 12, (s-5) % 12, s % 12]), tuple([(s - 9) % 12, (s-5) % 12, s % 12]), tuple([(s - 9) % 12, (s-6) % 12, s % 12])]
				else:
					addresses = [tuple([s, (s+3) % 12, (s+7) % 12])]
				values.append(sum([myDict.get(x, 0) for x in addresses]))
				#print t, values
			
			if not firstInversion:
				print("tonic", t, '\t'.join([str(round(x, 1)) for x in values]))
			plt.plot(values, label = str(t) + ['', '6'][int(firstInversion)])
	
	plt.xlabel('Root Note')
	plt.ylabel('Percentage of all sonorities with that root')
	ticks = range(len(theRoots))
	plt.xticks(ticks, [noteNameDict[x] for x in theRoots])
	legendTuple = []
	for t in tonalTypes:
		#s = noteNameDict[x]
		s = str(t)
		legendTuple.append(s + ' (root pos.)')
		legendTuple.append(s + ' (first inv.)')
	myLeg = plt.legend(tuple(legendTuple), loc=0)
	for i in range(len(myLeg.legendHandles)):
		myLeg.legendHandles[i].set_linewidth(2.0)
	plt.show()

def filter_tendencies(signature = None, tonic = None, inversions = 'F', reload = True, printOut = True):
	global approachDict, tendencyDict
	if reload:
		with open(DT._DATAPATH + _THISFILE + globalSource + 'tendencies.pkl', 'rb') as theFile:
			approachDict = pickle.load(theFile)
			tendencyDict = pickle.load(theFile)
	
	for myDict in [approachDict, tendencyDict]:
		delKeys = []
		for s in myDict:
			parsedS = parse_key(s)
			if parsedS[2][-1] in inversions:
				delKeys.append(s)
		for s in delKeys:
			del myDict[s]
	for myDict in [approachDict, tendencyDict]:
		delKeys = []
		for s in myDict:
			parsedS = parse_key(s)
			if parsedS[2][-1] in inversions:
				delKeys.append(s)
		for s in delKeys:
			del myDict[s]
	strSig1 = str(signature)
	strSig2 = str(tuple([signature]))
	for myDict in [approachDict, tendencyDict]:
		delKeys = []
		for s in myDict:
			parsedS = parse_key(s)
			if (tonic is not None) and parsedS[0] != tonic:
				delKeys.append(s)
			elif (signature is not None) and parsedS[1] not in [strSig1, strSig2]:
				delKeys.append(s)
		for s in delKeys:
			del myDict[s]
	if printOut:
		print("APPROACHED IN SPECIFIC WAYS")
		DT.print_dict(approachDict, cutoff = 20)
		print("FOLLOWED IN SPECIFIC WAYS")
		DT.print_dict(tendencyDict, cutoff = 20)

def multiple_tendencies(n = 3, myDict = None):
	global newTendencies, multiple_tendencies_n
	if myDict is None:
		myDict = approachDict
	newTendencies = {}
	multiple_tendencies_n = n
	for k in myDict:
		newKey = ' '.join(k.split(' ')[:-1])
		newTendencies.setdefault(newKey, []).append([k, myDict[k]])
	theKeys = sorted(newTendencies.keys(), key = newTendencies_sorting_key)
	for k, l in newTendencies.items():
		newTendencies[k] = sorted(l, key = lambda x: -x[1])
	for k in theKeys:
		print(k, newTendencies[k][:n], sum([newTendencies[k][i][1] for i in range(min(len(newTendencies[k]), n))])) 

def newTendencies_sorting_key(k):
	return -sum([newTendencies[k][i][1] for i in range(min(len(newTendencies[k]), multiple_tendencies_n))])

def parse_key(s):
	sSplit = s.split()
	return [sSplit[1], sSplit[3], sSplit[5], sSplit[6]]


def max_approached_by(mincount = 10, inversions = 'RTF', tonics = [0, 2, 4, 5, 7, 9, 11], targetSigs = [0, -1]):
	global approachDict
	approachDict = {}
	letterDict = {0: 'C', 2: 'D', 4: 'E', 5: 'F', 7: 'G', 9: 'A', 11: 'B'}
	for tonic in tonics:
		for targetSig in targetSigs:
			for letter in ['C', 'D', 'E', 'F', 'G', 'A', 'B']:
				if tonic == 4 and targetSig == -1:
					continue
				for inversion in inversions: #'RTF':  """RTF produces a lot of junk!"""
					approached_by(letter + inversion, tonic, targetSig, printOut = False)
					k = sorted([x for x in output.keys() if output[x] > mincount], key = lambda x:-output[x])[:5]
					totalSum = sum(output.values())
					prefix = 'tonic ' + letterDict[tonic] + ' signature ' + str(targetSig) + ' chord ' + letter + inversion + ' '
					print(prefix)
					for theKey in k:
						theValue = output[theKey] * 100./totalSum
						approachDict[prefix + theKey] = theValue
	tempTendencyDict = {}
	
	try:
		with open(DT._DATAPATH + _THISFILE + globalSource + 'tendencies.pkl', 'rb') as theFile:
			junkApproach = pickle.load(theFile)
			tempTendencyDict = pickle.load(theFile)
	except:
		tempTendencyDict = {}
	with open(DT._DATAPATH + _THISFILE + globalSource + 'tendencies.pkl', 'wb') as theFile:
		pickle.dump(approachDict, theFile)
		pickle.dump(tempTendencyDict, theFile)
	
	DT.print_dict(approachDict)

def greatest_tendency(mincount = 10, inversions = 'RTF', skipLength = 2, tonicList = [0, 2, 4, 5, 7, 9, 11], sigList = [0, -1]):
	global tendencyDict, output
	tendencyDict = {}
	letterDict = {0: 'C', 2: 'D', 4: 'E', 5: 'F', 7: 'G', 9: 'A', 11: 'B'}
	for tonic in tonicList:
		for targetSig in sigList:
			if type(targetSig) is int:
				trueSig = tuple([targetSig])
			else:
				trueSig = targetSig
			for letter in ['C', 'D', 'E', 'F', 'G', 'A', 'B']:
				if tonic == 4 and trueSig == (-1,):
					continue
				for inversion in inversions: #'RTF':  """RTF produces a lot of junk!"""
					limit_pieces(tuple([tonic]), trueSig, printOut = False)
					output = {}
					
					for i in bassProgs:
						if pieceList and i not in pieceList: continue
						for j in bassProgs[i]:
							if j > skipLength: continue
							for s, count in bassProgs[i][j].items():
								ind = 1
								if s[1] in ['-', '#']:
									ind += 1
								theNote = s[:ind]
								rawNote = theNote[0]
								theSym = s[ind:]
								theSym = theSym[:2] + theSym[4:]
				
								if theSym[-2] == '-':
									theInt = int(theSym[-2:])
								else:
									theInt = int(theSym[-1:])
								
								if theNote == letter and theSym[0] == inversion and theSym[1] in inversions:
					
									newSym = theNote + theSym
									output[newSym] = output.get(newSym, 0) + count
					
					k = sorted([x for x in output.keys() if output[x] > mincount], key = lambda x:-output[x])[:5]
					totalSum = sum(output.values())
					prefix = 'tonic ' + letterDict[tonic] + ' signature ' + str(targetSig) + ' chord ' + letter + inversion + ' '
					print(prefix)
					for theKey in k:
						theValue = output[theKey] * 100./totalSum
						tendencyDict[prefix + theKey] = theValue
					#return
	try:
		with open(DT._DATAPATH + _THISFILE + globalSource + 'tendencies.pkl', 'rb') as theFile:
			tempApproachDict = pickle.load(theFile)
			junkTendencyDict = pickle.load(theFile)
	except:
		tempApproachDict = {}
	with open(DT._DATAPATH + _THISFILE + globalSource + 'tendencies.pkl', 'wb+') as theFile:
		pickle.dump(tempApproachDict, theFile)
		pickle.dump(tendencyDict, theFile)
	DT.print_dict(tendencyDict)


def approached_by(targetChord = 'CR', tonic = 0, targetSig = 0, skipLength = 2, printOut = True):
	"""this is an older version, not so reliable; I pref"""
	global output
	
	letterNames = ['C', 'D', 'E', 'F', 'G', 'A', 'B']
	output = {}
	
	if type(targetSig) is int:
		targetSig = tuple([targetSig])
	
	limit_pieces(tuple([tonic]), targetSig, printOut = False)
	
	for i in bassProgs:
		if pieceList and i not in pieceList: continue
		for j in bassProgs[i]:
			if j > skipLength: continue
			for s, count in bassProgs[i][j].items():
				ind = 1
				if s[1] in ['-', '#']:
					ind += 1
				theNote = s[:ind]
				rawNote = theNote[0]
				theSym = s[ind:]
				theSym = theSym[:2] + theSym[4:]
				
				if theSym[-2] == '-':
					theInt = int(theSym[-2:])
				else:
					theInt = int(theSym[-1:])
					
				targetLetter = letterNames[(letterNames.index(rawNote) + theInt) % len(letterNames)]
				
				if targetLetter == targetChord[0] and theSym[1] == targetChord[1]:
					newSym = theNote + theSym
					output[newSym] = output.get(newSym, 0) + count
	if printOut:
		DT.print_dict(output, pct = True)

def approached_by_fifths():
	print("tonic\t", "\t".join(['C', 'D', 'F', 'G', 'A']))
	for m in [0, 2, 4, 7, 9]:
		v = []
		for c in 'CDFGA':
			approached_by(c + 'R', m, targetSig = 0, printOut = False)
			normalize_dict(output)
			l = [x[1] for x in output.items() if x[0].count('RR3') > 0]
			v.append(sum(l))
		print(m, "\t", '\t'.join([str(round(x, 1)) for x in v]))

currentPieceLimits = []
def limit_pieces(lastChord = (5,), theSig = (0,), bestChord = False, subsets = False, printOut = True):
	global pieceList, currentPieceLimits
	pieceList = []
	if type(lastChord) is int:
		lastChord = tuple([lastChord])
	if type(theSig) is int:
		theSig = tuple([theSig])
	if cd.globalSource != globalSource:
		cd.load(globalSource)
	currentPieceLimits = [lastChord[0], theSig[0]]
	cd.census(printOut = False)
	for k, l in cd.keyCensus.items():
		if subsets and bestChord:
			subsetMatch = all([x in bestChord for x in k[2]])
		if (not lastChord or k[0] == lastChord) and (not theSig or k[1] == theSig) and ((not bestChord) or (k[2] == bestChord or (subsets and subsetMatch))):
			if printOut:
				print(k, len(l))
			pieceList += l
	newPieceList = []
	for n in pieceList:
		name = cd.locNames[n]
		newPieceList.append(locNumbers[name])
	pieceList = newPieceList
	if printOut:
		print(len(pieceList), "pieces")

def accidentals(useDurations = False):
	global totalAccidentals
	totalAccidentals = {}
	if useDurations:
		targetDict = durationHisto
	else:
		targetDict = accidentalHisto
	for i in pieceList:
		for a, count in targetDict[i].items():
			totalAccidentals[a] = totalAccidentals.get(a, 0) + count

def B_natural_flat_ratio():
	global totalAccidentals
	for sig, tonicList in [[0, [0, 2, 4, 7, 9]], [-1, [0, 2, 5, 7, 9]]]:
		for tonic in tonicList:
			limit_pieces(tuple([tonic]), tuple([sig]), printOut = False)
			accidentals()
			letter = 'BE'[sig]
			b = totalAccidentals.get(letter, 0)
			bf = totalAccidentals.get(letter+'-', 0)
			print("sig, tonic, %", letter, "vs.", letter+'b', sig, tonic, b *100./max(b + bf, 1))

def how_many_root_progs():
	global remainder
	chord_progressions(2)
	totalProgs = {x[0]:len(x[1]) for x in longProgs.items()}
	rootProgs = {x[0]:len(x[1]) for x in longProgs.items() if x[0].count("R") == 2}
	thirdProgs = {x[0]:len(x[1]) for x in longProgs.items() if x[0].count("R") != 2 and x[0].count("T")}
	fifthProgs = {x[0]:len(x[1]) for x in longProgs.items() if x[0][1] == 'F' or x[0][2] == 'F' or x[0][-2] == 'F'}
	remainder = {x[0]:len(x[1]) for x in longProgs.items() if x[0] not in rootProgs and x[0] not in thirdProgs and x[0] not in fifthProgs}
	totalSum = sum(totalProgs.values())
	print('percent root-root progressions', round(sum(rootProgs.values())*100./totalSum,1))
	print('percent involving first inversion chords', round(sum(thirdProgs.values())*100./totalSum,1))
	print('percent involving second inversion chords', round(sum(fifthProgs.values())*100./totalSum,1))
	DT.print_dict(remainder, cutoff = 10)

def prog_matrix(ignoreMode = False, firstInv = False):
	chord_progressions(2)
	if firstInv:
		a = {x[0]:len(x[1]) for x in longProgs.items() if x[0].split()[0].count("T") and x[0].split()[1].count("R")}
	else:
		a = {x[0]:len(x[1]) for x in longProgs.items() if x[0].count("R") == 2}
	#column = 'CRM DRm DRM ERm FRM GRM ARm B-RM BRd'.split()
	#row = 'CRM DRm DRM ERm FRM GRM ARm B-RM BRd'.split()
	if ignoreMode:
		column = 'C D E F G A B- B'.split()
	else:
		column = 'CRM DRm DRM ERm ERM FRM GRm GRM ARm ARM B-RM'.split()
	row = column[:]
	
	totalSum = 0
	matrix = []
	for c in column:
		out = []
		if firstInv and ignoreMode:
			c+='T'
		else:
			c+='R'
		for d in row:
			if firstInv and ignoreMode:
				d+='R'
			else:
				d+='R'
			if ignoreMode:
				s = 0
				for s1 in 'Mmd':
					for s2 in 'Mmd':
						s += a.get(c + s1 + ' ' + d + s2, 0)
			else:
				s = a.get(c + ' ' + d, 0)
			totalSum += s
			out.append(s)
		matrix.append(out)
	print('\t' + '\t'.join(column))
	for i, values in enumerate(matrix):
		newV = [str(round(x*100./totalSum, 1)) for x in values]
		print(row[i] + '\t' + '\t'.join(newV))

def accidental_histo(lastChord = 0, theSig = 0, bestChord = None, letterNames = 'FCGDAEB', useDurations = False):
	limit_pieces(lastChord = lastChord, theSig = theSig, bestChord = bestChord)
	accidentals(useDurations = useDurations)
	
	baseAccis = [totalAccidentals.get(x, 0) for x in letterNames]
	alterations = [totalAccidentals.get(x + '#', 0) + totalAccidentals.get(x + '-', 0) for x in letterNames]
	
	theSum = (sum(baseAccis) + sum(alterations))/100.
	
	baseAccis = [x/theSum for x in baseAccis]
	alterations = [x/theSum for x in alterations]
	
	#print baseAccis
	#print 

	p1 = plt.bar(range(7), baseAccis, .35, color = 'black')
	p2 = plt.bar(range(7), alterations, .35, bottom=baseAccis, color = 'grey')
	
	modeNames = {0: 'ionian', 2: 'dorian', 4: 'phrygian', 7: 'mixolydian', 9: 'aeolian'}
	
	if theSig == -1:
		modeNames = {(x[0] - 7) % 12:x[1] for x in modeNames.items()[:]}
	
	plt.ylabel('Note Percentage')
	plt.title('Note distributions for ' + modeNames[lastChord] + ' mode')
	plt.xticks(range(7), tuple(letterNames))
	plt.legend((p1[0], p2[0]), ('Unaltered', 'Altered'))

	plt.show()

			
def compare_frequencies(sym1 = 'ER', sym2 = 'ET', ignoreQualities = True, sig = 0, tonicList = [0, 2, 4, 7, 9], printOut = True):
	global totalAccidentals
	averageList = []
	for tonic in tonicList:
		limit_pieces(tuple([tonic]), tuple([sig]), printOut = False)
		chord_progressions(1)
		if ignoreQualities:
			b = 0
			bf = 0
			for k, l in longProgs.items():
				if k.count("?") != 0: continue
				if k.startswith(sym1):
					b += len(l)
				elif k.startswith(sym2):
					bf += len(l)
		else:
			b = len(longProgs.get(sym1, []))
			bf = len(longProgs.get(sym2, []))
		avg = b *100./max(b + bf, 1)
		averageList.append(avg)
		if printOut:
			print("sig, tonic, %", sig, tonic, int(round(avg, 0)))
	if printOut:
		print("AVERAGE", round(sum(averageList)/float(len(averageList)), 1))
	return averageList

def compare_qualities(qualities = 'RR', interval = '3', qual = ['M', 'm']):
	global counts
	counts = [	sum([progs[x] for x in progs.keys() if (len(x) >= 5 and x[0] == qualities[0][0] and x[2] == qual[0])]), 
			sum([progs[x] for x in progs.keys() if (len(x) >= 5 and x[0] == qualities[0][0] and x[2] == qual[1])])]
	for i in [0, 1]:
		for sign in ['', '-']:
			found = 0
			for destQual in ['M', 'm']:
				key = qualities + qual[i] + destQual + sign + interval
				temp = progs.get(key, 0)
				found += temp
				print("  ", key, temp)
			print("TOTAL", round(found * 100./counts[i], 1))
			

def normalize_dict(myDict):
	theSum = sum(myDict.values())
	for k, v in myDict.items():
		myDict[k] = v * 100./theSum

def piece_report(pieceNumber):
	global pieceList, pongProgs
	pieceList = [pieceNumber]
	chord_progressions(1)
	m = sorted(longProgs.keys(), key = lambda x: -len(longProgs[x]))
	for i in m: print(i, len(longProgs[i]))

def report(myString, skipLength = 2):
	global output, pieceList
	output = {}
	totalBassProgs = {}
	for i in bassProgs:
		if pieceList and i not in pieceList: continue
		for j in bassProgs[i]:
			if j > skipLength: continue
			for s, count in bassProgs[i][j].items():
				ind = 1
				if s[1] in ['-', '#']:
					ind += 1
				theNote = s[:ind]
				totalBassProgs[theNote] = totalBassProgs.get(theNote, 0) + count
				theSym = s[ind:]
				theSym = theSym[:2] + theSym[4:]
				if theSym == myString:
					output[theNote] = output.get(theNote, 0) + count
	
	for s in output:
		output[s] = output[s] * 100./totalBassProgs[s]
	
	DT.print_dict(output)

#pieceList = []

def pachelbel(skipLength = 3, minLength = 3, patterns = [[4, 1], [1, 4]]):
	global output, pieceList
	if not pieceList:
		pieceList = fullProgs.keys()
	for p in pieceList:
		fullProg = fullProgs[p]
		chordString = ' '.join([x[1] for x in fullProg])
		for i in range(skipLength):
			chordString.replace('? ', '')
		trimmedString = [get_root(x) for x in chordString.split()]
		intervalList = make_interval_list(trimmedString)
		#intervalList = [x for x in intervalList if x != 0]
		"""print chordString
		print ' '.join([str(x) for x in trimmedString])
		print intervalList"""
		maxLens = [0]
		for pattern in patterns:
			patternLoc, lengths = max_patternlength(intervalList, pattern)
			maxLens += lengths
		m = max(maxLens)
		if m >= minLength: print("piece", p, "max length", m)

def make_interval_list(l):
	outList = []
	for i in range(len(l) - 1):
		if type(l[i]) is int and type(l[i+1]) is int:
			outList.append((l[i+1] - l[i]) % 7)
		elif outList and outList[-1] != -1:
			outList.append(-1)
	return outList
	
def max_patternlength(l, pattern):
	patternLength = len(pattern)
	patternLocations = [i for i in range(len(l) - patternLength + 1) if l[i:i+patternLength] == pattern]
	repeatNumbers = []
	for i, curLoc in enumerate(patternLocations):
		instances = 1
		while curLoc + patternLength in patternLocations:
			instances += 1
			curLoc += patternLength
		repeatNumbers.append(instances)
	return patternLocations, repeatNumbers


def get_root(s):
	if s[0] not in globalLetterNames: 
		return '?'
	sd = globalLetterNames.index(s[0])
	if s.count("T") > 0 or s.count('6') > 0:
		sd = sd - 2
	elif s[1:].count("F") > 0:
		sd = sd - 4
	return sd % 7

def get_thirds(s):
	if s[0] not in globalLetterNames: 
		return '?'
	sd = globalLetterNames.index(s[0])
	if s.count("R") > 0:
		return [sd % 7, (sd + 2) % 7]
	elif s.count("3") > 0:
		return [sd % 7]
	elif s.count("T") > 0:
		return [(sd - 2) % 7, sd % 7]
	elif s.count('6') > 0:
		return [(sd - 2) % 7]
	elif s[1:].count("F") > 0:
		return [(sd - 4) % 7, (sd - 2) % 7]
	return False

def intString(s):
	output = []
	for i in range(len(s) - 1):
		tempLetter = get_interval_from_letters

def chain_of_thirds(targetInt = 6, allowableReps = 0):
	for k in longProgs:
		chords = k.split()
		thirds = [get_thirds(x) for x in chords]
		if not thirds[0]: continue
		lastThirds = thirds[0][:]
		thirdList = [lastThirds[:]]
		for t in thirds[1:]:
			if not lastThirds: break
			if not t: 
				lastThirds = False
				break
			newThirds = []
			for test in t:
				if test in lastThirds or (test - targetInt) % 7 in lastThirds:
					newThirds.append(test)
			lastThirds = newThirds[:]
			thirdList.append(lastThirds[:])
		if lastThirds:
			if len(thirdList[-1]) == 2: continue
			finalThird = thirdList[-1][0]
			totalReps = 0
			for t in thirdList[-2::-1]:
				if finalThird in t:
					totalReps += 1
				else:
					finalThird = (finalThird - targetInt) % 7
			if totalReps <= allowableReps:
				print(k, thirdList, len(longProgs[k]))

def pachelbel(sequenceList = [4, 1]):
	for k in longProgs:
		chords = k.split()
		roots = [get_root(x) for x in chords]
		rootInts = [(roots[i] - roots[i-1]) % 7 for i in range(1, len(roots))]
		poss1 = sequenceList * len(rootInts)
		poss2 = sequenceList[::-1] * len(rootInts)
		if rootInts == poss1[:len(rootInts)] or rootInts == poss2[:len(rootInts)]:
			print(k, roots, len(longProgs[k]))

def string_of_fifths(targetInt = 4):
	for k in longProgs:
		chords = k.split()
		roots = [get_root(x) for x in chords]
		stackOfFifths = [(roots[0] + (x * targetInt)) % 7 for x in range(len(roots))]
		if stackOfFifths == roots:
			print(k, roots, len(longProgs[k]))

triadNames = {'M': [0, 4, 7], 'm': [0, 3, 7], 'd': [0, 3, 6], 'A': [0, 4, 8], '+': [0, 4, 8]}
incompleteSonorities = {'1P': [0], '3M': [0, 4], '3m': [0, 3], '5P': [0, 7], '5d': [0, 6], '6M': [0, 9], '6m': [0, 8]}
globalPitchClasses = {'C': 0, 'D': 2, 'E': 4, 'F': 5, 'G': 7, 'A': 9, 'B': 11}

def subset(s, t):
	try:
		return set(parse_figure(s)) <= set(parse_figure(t))
	except:
		return False

def parse_figure(s):
	if s[0] not in globalPitchClasses: 
		return False
	sd = globalPitchClasses[s[0]]
	sd = sd + s.count("#") - s.count('-')
	remainingFig = s[1:].replace('#', '').replace('-', '')
	if remainingFig[0] in 'RTF':
		triad = triadNames.get(remainingFig[1], False)
		if not triad:
			print("ERROR HERE", s)
			return False
		inversion = 'RTF'.index(remainingFig[0])
		triad = triad[inversion:] + triad[:inversion]
		offset = (sd - triad[0]) % 12
		triad = [(x + offset) % 12 for x in triad]
		return triad
	else:
		chord = incompleteSonorities.get(remainingFig, False)
		if not chord:
			print("ERROR HERE", s)
			return False
		offset = (sd - chord[0]) % 12
		chord = [(x + offset) % 12 for x in chord]
		return chord
	return False

def print_piece_list():
	global pieceList
	
def translate_piece_numbers(oldNum, oldLocNames):
	baseName = oldLocNames[oldNum]
	return locNumbers[basName]

def last_chord_by_piece(printOut = True, matchSig = None, usePieceList = False):
	global massFiles, finalChords, keySigs
	#def limit_pieces(lastChord = (5,), theSig = (0,), bestChord = False, printOut = True):
	l = list(set(([x[0] for x in cd.palestrinaData.values()])))
	massFiles = {}
	for x in cd.palestrinaData.items():
		massFiles.setdefault(x[1][0], []).append(x[0])
	
	finalChords = {x:[] for x in massFiles.keys()}
	keySigs = {x:[] for x in massFiles.keys()}
	
	cd.census(printOut = False)
	
	for k, l in cd.keyCensus.items():
		for pieceNumber in l:
			if usePieceList:
				if pieceNumber not in pieceList: continue
			fileName = cd.locNames[pieceNumber]
			massName = cd.palestrinaData[fileName][0]
			finalChords[massName].append([fileName, k[0]])
			keySigs[massName].append(k[1])
	
	totalNames = set([])
		
	if printOut:
		for i in sorted(finalChords.keys()):
			s = sorted(finalChords[i], key = sort_mass_movements)
			if not s: continue
			out = []
			for name, f in s:
				baseName = name.split('_')[0]
				totalNames.add(baseName)
				out.append(baseName + ' ' + str(f)) 
			if len(set(keySigs[i])) > 1:
				print("MULTIPLE KEY SIGNATURES", i, keySigs[i])
				print(i, keySigs[i][0])
				print('   ', ', '.join(out))
			elif (matchSig is None) or (keySigs[i][0] == matchSig):
				print(i, keySigs[i][0])
				print('   ', ', '.join(out))
				
	#print totalNames

def sort_mass_movements(theList):
	order = [['K', 0], ['G', 1], ['C', 2], ['S', 3], ['B', 4], ['A', 5]]
	for s, i in order:
		if theList[0].startswith(s):
			return i
	return 6
	
def get_piece_files(s = None, n = None, printOut = True):
	global massFiles
	massFiles = {}
	for x in cd.palestrinaData.items():
		massFiles.setdefault(x[1][0], []).append(x[0])
	l = sorted(massFiles.get(s, []), key = sort_mass_movements)
	if printOut:
		print(l)
		if n is not None:
			corpus.parse(l[n]).show()

"""
======================================================================
new code written for TAOM, probably more reliable as it uses longProgs
======================================================================
"""

def fig641():
	global longProgs
	chord_progressions(1)
	longProgs = {x[0]:len(x[1]) for x in longProgs.items() if x[0][1] in "RTF" and x[0][2] in "AMmd"}
	totalChords = sum(longProgs.values())
	print('major 5/3', int(round(sum([x[1] for x in longProgs.items() if x[0][1:] == 'RM'])*100./totalChords, 0)))
	print('major 6/3', int(round(sum([x[1] for x in longProgs.items() if x[0][1:] == 'TM'])*100./totalChords, 0)))
	print('minor 5/3', int(round(sum([x[1] for x in longProgs.items() if x[0][1:] == 'Rm'])*100./totalChords, 0)))
	print('minor 6/3', int(round(sum([x[1] for x in longProgs.items() if x[0][1:] == 'Tm'])*100./totalChords, 0)))
	print('dim 5/3', int(round(sum([x[1] for x in longProgs.items() if x[0][1:] == 'Rd'])*100./totalChords, 0)))
	print('dim 6/3', int(round(sum([x[1] for x in longProgs.items() if x[0][1:] == 'Td'])*100./totalChords, 0)))
	print('all 6/4', int(round(sum([x[1] for x in longProgs.items() if x[0][1] == 'F'])*100./totalChords, 0)))

def fig863(cutoff = 4):
	global longProgs, output
	load("palraw")
	modeInfo = [(0, 0, (0, 4, 7)), (7, 0, (7, 11, 2)), (2, 0, (2, 5, 9)), (9, 0, (9, 0, 4))]
	modeNames = ['ionian', 'mixolydian', 'dorian', 'aeolian']
	combineChords = {1: [['V', 'v'], ['I', 'i']], 2: [['V', 'v'], ['I', 'i'], ['IV', 'iv']], 3: [['V', 'v'], ['I', 'i'], ['IV', 'iv']]}
	for i, m in enumerate(modeInfo):
		print(modeNames[i])
		limit_pieces(*m)
		chord_progressions(1)
		output = {translate_symbols(x[0]):len(x[1]) for x in longProgs.items()}
		cc = combineChords.get(i, [])
		for cList in cc:
			baseChord = ' or '.join(cList)
			output[baseChord] = 0
			for c in cList:
				output[baseChord] = output[baseChord] + output.get(c, 0)
				del output[c]
		output = convert_dict_to_percentages(output)
		for k, pct in sorted(output.items(), key = lambda x: -x[1]):
			if pct > cutoff:
				print(k, pct)

def convert_dict_to_percentages(d, digits = 0):
	s = sum(d.values())
	if digits == 0:
		return {x[0]:int(round(x[1]*100./s, 0)) for x in d.items()}
	else:
		return {x[0]:round(x[1]*100./s, digits) for x in d.items()}

def tendency_summary(minCount = 0, thresh = 15, **kwargs):
	finalDict = {}
	for targetChord in ['CRM', 'DRm', 'ERm', 'FRM', 'GRM', 'ARm', 'BRd', 'CTm', 'DTd', 'ETM', 'FTm', 'GTm', 'ATM', 'BTM']:
		d = new_tendency(targetChord, printOut = False, **kwargs)
		d = {x[0]:x[1] for x in d.items() if x[1] > minCount}
		d = convert_dict_to_percentages(d)
		finalDict.update(d)
	output = sorted(finalDict.items(), key = lambda x: -x[1])
	for pair in output:
		if pair[1] > thresh:
			print(pair[0].ljust(12) + '\t', pair[1])

def new_distributions(progLength = 1, tonic = 0, targetSig = 0, bestChord = False, printOut = True):
	global output
	output = {}
	
	if type(targetSig) is int:
		targetSig = tuple([targetSig])
	
	limit_pieces(tuple([tonic]), targetSig, bestChord = bestChord, printOut = False)
	chord_progressions(progLength)
	
	
	output = {' '.join([translate_symbol(y) for y in x[0].split()]):len(x[1]) for x in longProgs.items()}
	if printOut: DT.print_dict(output, pct = True)
	return output

def get_all_modes(modeType = 'major', keySigs = [0, -1], subsets = True):
	global pieceList
	totalPieceList = []
	for keySig in keySigs:
		theScale = [0, 2, 4, 5, 7, 9, 11]
		if keySig == -1:
			theScale[-1] = 10
		for i, tonic in enumerate(theScale):
			if modeType == 'major' and (theScale[(i+2)%7] - tonic) % 12 != 4:
				continue
			if modeType == 'minor' and (theScale[(i+2)%7] - tonic) % 12 != 3:
				continue
			for j, bestFinal in enumerate(theScale):				
				triad = (bestFinal, theScale[(j+2)%7], theScale[(j+4)%7])
				if modeType == 'major' and (triad[1] - triad[0]) % 12 != 4:
					continue
				if modeType == 'minor' and (triad[1] - triad[0]) % 12 != 3:
					continue
				limit_pieces(tonic, keySig, bestChord = triad, printOut = False, subsets = subsets)
				totalPieceList += pieceList
	pieceList = totalPieceList
	#print(len(pieceList), 'total pieces')
		
def count_chord_type(targetType = 'M', targetMode = '', includeIncomplete = True):
	global longProgs
	chord_progressions(1, includeIncomplete = includeIncomplete)
	longProgs = {x[0]:len(x[1]) for x in longProgs.items() if x[0][-1] in 'AMmd'}
	totalSum = sum(longProgs.values())
	if includeIncomplete:
		foundSum = sum([x[1] for x in longProgs.items() if x[0].count(targetType) and x[0][1] in 'RTF'])
	else:
		foundSum = sum([x[1] for x in longProgs.items() if x[0].count(targetType)])
	print('  ', int(round(100.*foundSum/totalSum, 0)), 'percent matches', targetType)

def summarize_chord_types(composerList = ['Ockeghem', 'Josquincomplete', 'Palraw'], includeIncomplete = True):
	global pieceList
	for c in composerList:
		print("\n=====", c, "=====")
		load(c)
		get_all_modes('major')
		print(' major modes', len(pieceList), 'total pieces')
		print('  major chords')
		count_chord_type('M', includeIncomplete = includeIncomplete)
		print('  minor chords')
		count_chord_type('m', includeIncomplete = includeIncomplete)
		get_all_modes('minor')
		print('\n minor modes', len(pieceList), 'total pieces')
		print('  major chords')
		count_chord_type('M', includeIncomplete = includeIncomplete)
		print('  minor chords')
		count_chord_type('m', includeIncomplete = includeIncomplete)	

def new_tendency(targetChord = 'CRM', chordPosition = 0, tonic = 0, targetSig = 0, skipLength = 2, printOut = True):
	"""
	uses longprogs to find chords that are approached by, or go to, other chords particularly frequently
	
	to find chords that move in a particular way, set chordPosition == 0
	to find chords approached in a particular way, set chordPosition == 1
	
	"""
	global output
	
	letterNames = ['C', 'D', 'E', 'F', 'G', 'A', 'B']
	output = {}
	
	if type(targetSig) is int:
		targetSig = tuple([targetSig])
	
	limit_pieces(tuple([tonic]), targetSig, printOut = False)
	chord_progressions(2)
	
	if chordPosition == 0:
		findTarget = targetChord + ' '
	elif chordPosition == 1:
		findTarget = ' ' + targetChord
	
	output = {' '.join([translate_symbol(y) for y in x[0].split()]):len(x[1]) for x in longProgs.items() if x[0].count(findTarget)}
	if printOut: DT.print_dict(output, pct = True)
	return output

def translate_symbol(s, tonic = None, keySig = None):
	"""needs to be made to work for 1-flat key signaturs"""
	global currentPieceLimits
	if tonic is None and currentPieceLimits:
		tonic, keySig = currentPieceLimits
	else:
		tonic = 0
		keySig = 0
	if type(keySig) is tuple:
		keySig = keySig[0]
	if type(tonic) is tuple:
		tonic = tonic[0]
	originalS = s
	letters = 'CDEFGAB'
	diatonicLabels = {0: 'C', 2: 'D', 4: 'E', 5: 'F', 7: 'G', 9: 'A', 11: 'B'}
	tonicAdjustment = 0
	if keySig == -1:
		s = s.replace('B-', 'B')
	elif keySig == 1:
		s = s.replace('F#', 'F')
	elif keySig == 2:
		s = s.replace('C#', 'C').replace('F#', 'F')
	if tonic:
		tonicAdjustment = -letters.index(diatonicLabels.get(tonic, 0))
	rns = {0: 'I', 1: 'II', 2: 'III', 3: 'IV', 4: 'V', 5: 'VI', 6: 'VII'}
	if s[0] not in letters:
		return None
	root = letters.index(s[0])
	inv = s[-2]
	invSym = ''
	if inv == 'T':
		root = root - 2
		invSym = '6'
	elif inv == 'F':
		root = root - 4
		invSym = '6/4'
	out = rns.get((root + tonicAdjustment) % 7, None)
	if s[-1] == 'M':
		out = out.upper()
	elif s[-1] == 'm':
		out = out.lower()
	elif s[-1] == 'd':
		out = out.lower() + 'o'	
	elif s[-1] == 'A':
		out = out.upper() + '+'
	out += invSym	
	if s[1] == '-':
		out = 'b'+out
	if s[1] == '+':
		out = '#'+out
	if s.count('?'): out = out+'?'
	if out == 'bv6': out = 'v6'
	if out == 'ivo6': out = '#ivo6'
	return out

def translate_symbols(s):
	return ' '.join([translate_symbol(x) for x in s.split()])

def average_ionian_mode(progLength = 1):
	global cMajor, fMajor
	cMajor = new_distributions(progLength = progLength, printOut = False, tonic = 0, targetSig = 0)#, bestChord = (0, 4, 7))
	fMajor = new_distributions(progLength = progLength, printOut = False, tonic = 5, targetSig = -1)#, bestChord = (5, 9, 0))
	if globalSource == 'VICTORIARAW':
		gMajor = new_distributions(progLength = progLength, printOut = False, tonic = 7, targetSig = 1)#, bestChord = (7, 11, 2))
		dMajor = new_distributions(progLength = progLength, printOut = False, tonic = 2, targetSig = 2)#, bestChord = (2, 6, 9))
	totalKeys = list(set(list(cMajor.keys()) + list(fMajor.keys())))
	outDict = {}
	for k in totalKeys:
		outDict[k] = cMajor.get(k, 0) + fMajor.get(k, 0)
		if globalSource == 'VICTORIARAW':
			outDict[k] += fMajor.get(k, 0) + gMajor.get(k, 0)
	return outDict
	
def chord_histo(composers = ['Palraw', 'VictoriaRaw', 'Ockeghem'], filterRare = 1, averageIonian = True, **kwargs):
	global totalDict
	fig = plt.figure()
	ax = fig.add_subplot(1, 1, 1)
	
	replacementNames = {'JOSQUINCOMPLETE': 'Josquin', 'PALRAW': "Palestrina", "VICTORIARAW": "Victoria"}
	
	totalDict = []
	for c in composers:
		load(c)
		if averageIonian:
			graphItems = average_ionian_mode()
			graphItems = convert_dict_to_percentages(graphItems, 2)
		else:
			graphItems = new_distributions(printOut = False, **kwargs)
			graphItems = convert_dict_to_percentages(graphItems, digits = 2)
		totalDict.append(graphItems)
	
	totalKeys = []
	for d in totalDict:
		totalKeys += d.keys()
	
	totalKeys = list(set(totalKeys))
	totalValues = {}
	for k in totalKeys:
		totalValues[k] = [d.get(k, 0) for d in  totalDict]
	
	finalKeys = [k for k in totalKeys if max(totalValues[k]) > filterRare]
	
	finalKeys = sorted(finalKeys, key = lambda x: -totalValues[x][0])
	
	yMax = 0
	for i, c in enumerate(composers):
		newList = [totalValues[k][i] for k in finalKeys]
		yMax = max(max(newList), yMax)
		print(c, [str(finalKeys[i]) + ' ' + str(round(newList[i], 2)) for i in range(len(newList))])
		plt.plot(newList, label=c)
	
	labels = [''.join([str(x) for x in y]) for y in finalKeys]
	
	#fig, ax = plt.subplots()
	
	plt.xlabel('Roman numeral')
	plt.ylabel('Percentage of all complete triads')
	ticks = range(len(labels))
	#plt.xticks(rotation=-90)
	plt.xticks(ticks, labels)
	ax.set_xlim(0, len(labels) -1)
	#xMax = len(ticks)
	#ax.set_xlim(0, xMax)
	tempComps = [replacementNames[x.upper()] if x.upper() in replacementNames else x for x in composers]
	
	myLeg = plt.legend(tuple(tempComps), loc=0)
	for i in range(len(myLeg.legendHandles)):
		myLeg.legendHandles[i].set_linewidth(2.0)
	plt.show()

def tendency_histo(progStr = 'IV -> ', removeOriginal = True, averageIonian = True, sortByTendency = True, cutDifference = .1, **kwargs):
	global progDict, longProgs
	fig = plt.figure()
	ax = fig.add_subplot(1, 1, 1)
	composer = globalSource[0] + globalSource[1:].lower()
	myChord = ''
	if progStr[-1] == '>': 
		progStr += ' '
	elif progStr[0] == '-':
		progStr = ' ' + progStr
	if progStr[-4:] == ' -> ':
		myChord = progStr.split(' -> ')[-2]
		endVar = True
	elif progStr[:4] == ' -> ':
		endVar = False
		myChord = progStr.split(' -> ')[1]
	else:
		print("INVALID PROGSTR, SHOULD START OR END WITH ' -> '")
		return
	
	styleStrings = ['-', '--', '-','--', '-', '--', '-','--']
	widths = [1, 2, 1, 2, 1, 2, 1, 2]
	markers = ['', '', '', '', '', '', '', '', '', '']
	plt.title('Histogram of Chord Tendencies in ' + composer, fontsize = 24)
	plt.ylabel('Percentage', fontsize = 18)
	if endVar:
		plt.xlabel('Chords that follow ' + progStr.rstrip(' ->'), fontsize = 18)
	else:
		plt.xlabel('Chords that precede ' + progStr.lstrip(' ->'), fontsize = 18)
	
	values = []
	labels = []
	values2 = []
	
	if averageIonian:
		progDict = average_ionian_mode(2)
	else:
		progDict = new_distributions(2, printOut = False, **kwargs)
	
	if endVar:
		progDict = {x[0]:x[1] for x in progDict.items() if x[0].startswith(myChord + ' ')}
	else:
		progDict = {x[0]:x[1] for x in progDict.items() if x[0].endswith(' ' + myChord)}
	
	totalCounts = sum(progDict.values())
	print("total two-chord progressions", totalCounts)
	
	for item in progDict.items():
		values.append(100.*item[1]/totalCounts)
		if endVar:
			labels.append(item[0].split()[1])												# change here
		else:
			labels.append(item[0].split()[0])
	
	if averageIonian:
		longProgs = average_ionian_mode(1)
	else:
		longProgs = new_distributions(1, printOut = False, **kwargs)
	
	if removeOriginal and myChord and myChord in longProgs:
		del longProgs[myChord]
	
	for item in labels:
		if item in longProgs:
			values2.append(longProgs[item])
		else:
			values2.append(0)
	
	print("total one-chord counts", sum(longProgs.values()))
	values2 = [100. * x / sum(longProgs.values()) for x in values2]	# or sum of values2?
	
	if sortByTendency:
		indexes = range(len(values))
		indexes = sorted(indexes, key = lambda x: values2[x] - values[x])
		newValues = []
		newValues2 = []
		newLabels = []
		for i in indexes:
			if abs(values2[i] - values[i]) < cutDifference:						# for readability skip things whose tendency is essentially 0
				continue
			newValues.append(values[i])
			newValues2.append(values2[i])
			newLabels.append(labels[i])
		values = newValues
		values2 = newValues2
		labels = newLabels
	plt.plot(values, label=progStr, linestyle=styleStrings[0], marker = markers[0], linewidth = widths[0])
	plt.plot(values2, label='raw probs.', linestyle=styleStrings[1], marker = markers[1], linewidth = widths[1])
	ticks = range(len(labels))
	mySize = 18
	if len(labels) > 25:
		mySize = 14
	plt.xticks(ticks, labels)
	plt.xticks(rotation=90)
	for i in ticks:
		print(labels[i], "%.1f" % (values[i] - values2[i]))
	myLeg = plt.legend((progStr, 'raw probs.'), loc=0)
	for i in range(len(myLeg.legendHandles)):
		myLeg.legendHandles[i].set_linewidth(2.0)
	ax.set_xlim(0, len(labels) -1)
	plt.show()
